"""Utilities to retrieve package metadata."""

from __future__ import annotations

from email.message import Message
from importlib.metadata import metadata
from typing import cast

from ._models import Metadata

__all__ = [
    "get_metadata",
    "get_project_url",
]


def get_metadata(*, package_name: str, application_name: str) -> Metadata:
    """Retrieve metadata for the application.

    Parameters
    ----------
    pacakge_name
        The name of the package (Python namespace). This name is used to look
        up metadata about the package.
    application_name
        The value to return as the application name (the ``name`` metadata
        field).

    Returns
    -------
    Metadata
        The package metadata as a Pydantic model, suitable for returning as
        the result of a FastAPI route.

    Notes
    -----
    ``get_metadata`` integrates extensively with your package's metadata.
    Typically this metadata is set in ``pyproject.toml``, ``setup.cfg``,
    ``setup.py`` file (for setuptools-based applications).  The
    ``pyproject.toml`` fields used are:

    version
        Used as the version metadata. This may be set automatically with
        ``setuptools_scm``.
    description
        Use as the ``description`` metadata.
    project.urls, Homepage
        Used as the ``documentation_url`` metadata.
    project.urls, Source
        Used as the ``respository_url``.

    Packages using ``setup.cfg`` or ``setup.py`` get the last three items of
    metadata from different sources:

    summary
        Use as the ``description`` metadata.
    url
        Used as the ``documentation_url`` metadata.
    project_urls, Source code
        Used as the ``respository_url``.
    """
    pkg_metadata = cast("Message", metadata(package_name))

    # Newer packages that use pyproject.toml only do not use the Home-page
    # field (setuptools in pyproject.toml mode does not support it) and use
    # different names for the project URLs.  Attempt those names first and
    # fall back to the older names.
    repository_url = get_project_url(pkg_metadata, "Source")
    if not repository_url:
        repository_url = get_project_url(pkg_metadata, "Source code")
    documentation_url = get_project_url(pkg_metadata, "Homepage")
    if not documentation_url:
        documentation_url = pkg_metadata.get("Home-page", None)

    return Metadata(
        name=application_name,
        version=pkg_metadata.get("Version", "0.0.0"),
        description=pkg_metadata.get("Summary", None),
        repository_url=repository_url,
        documentation_url=documentation_url,
    )


def get_project_url(meta: Message, label: str) -> str | None:
    """Get a specific URL from a package's ``project_urls`` metadata.

    Parameters
    ----------
    meta
        The package metadata, as returned by the
        ``importlib.metadata.metadata`` function.
    label
        The URL's label. Consider the follow snippet of a ``pyproject.toml``
        file:

        .. code-block:: toml

           [project.urls]
           Homepage = "https://safir.lsst.io/"
           Source = "https://github.com/lsst-sqre/safir"

        To get the ``https://github.com/lsst-sqre/safir`` URL, the label is
        ``Source``.

        Packages using ``setup.cfg`` use a different syntax but a similar
        approach.

        .. code-block:: ini

           project_urls =
               Change log = https://safir.lsst.io/changelog.html
               Source code = https://github.com/lsst-sqre/safir
               Issue tracker = https://github.com/lsst-sqre/safir/issues

        To get the ``https://github.com/lsst-sqre/safir`` URL, the label is
        ``Source code``.

    Returns
    -------
    str or None
        The URL. If the label is not found, the function returns `None`.

    Examples
    --------
    >>> from importlib_metadata import metadata
    >>> meta = metadata("safir")
    >>> get_project_url(meta, "Source")
    'https://github.com/lsst-sqre/safir'
    """
    prefix = f"{label}, "
    for key, value in meta.items():
        if key == "Project-URL":
            if value.startswith(prefix):
                return value[len(prefix) :]
    return None



"""Tests for the job list.

These tests don't assume any given application, and therefore don't use the
API to create a job, instead inserting it directly via the UWSService.
"""

from __future__ import annotations

from datetime import timedelta

import pytest
from httpx import AsyncClient
from vo_models.uws import Jobs

from safir.datetime import current_datetime, isodatetime
from safir.testing.uws import MockWobbly
from safir.uws._dependencies import UWSFactory

from ..support.uws import SimpleParameters

FULL_JOB_LIST = """
<uws:jobs
    version="1.1"
    xsi:schemaLocation="http://www.ivoa.net/xml/UWS/v1.0 UWS.xsd"
    xmlns:xml="http://www.w3.org/XML/1998/namespace"
    xmlns:uws="http://www.ivoa.net/xml/UWS/v1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <uws:jobref id="3" xlink:href="https://example.com/test/jobs/3">
    <uws:phase>PENDING</uws:phase>
    <uws:ownerId>test-user</uws:ownerId>
    <uws:creationTime>{}</uws:creationTime>
  </uws:jobref>
  <uws:jobref id="2" xlink:href="https://example.com/test/jobs/2">
    <uws:phase>PENDING</uws:phase>
    <uws:runId>some-run-id</uws:runId>
    <uws:ownerId>test-user</uws:ownerId>
    <uws:creationTime>{}</uws:creationTime>
  </uws:jobref>
  <uws:jobref id="1" xlink:href="https://example.com/test/jobs/1">
    <uws:phase>PENDING</uws:phase>
    <uws:ownerId>test-user</uws:ownerId>
    <uws:creationTime>{}</uws:creationTime>
  </uws:jobref>
</uws:jobs>
"""

RECENT_JOB_LIST = """
<uws:jobs
    version="1.1"
    xsi:schemaLocation="http://www.ivoa.net/xml/UWS/v1.0 UWS.xsd"
    xmlns:xml="http://www.w3.org/XML/1998/namespace"
    xmlns:uws="http://www.ivoa.net/xml/UWS/v1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <uws:jobref id="3" xlink:href="https://example.com/test/jobs/3">
    <uws:phase>PENDING</uws:phase>
    <uws:ownerId>test-user</uws:ownerId>
    <uws:creationTime>{}</uws:creationTime>
  </uws:jobref>
</uws:jobs>
"""

QUEUED_JOB_LIST = """
<uws:jobs
    version="1.1"
    xsi:schemaLocation="http://www.ivoa.net/xml/UWS/v1.0 UWS.xsd"
    xmlns:xml="http://www.w3.org/XML/1998/namespace"
    xmlns:uws="http://www.ivoa.net/xml/UWS/v1.0"
    xmlns:xlink="http://www.w3.org/1999/xlink"
    xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance">
  <uws:jobref id="2" xlink:href="https://example.com/test/jobs/2">
    <uws:phase>QUEUED</uws:phase>
    <uws:runId>some-run-id</uws:runId>
    <uws:ownerId>test-user</uws:ownerId>
    <uws:creationTime>{}</uws:creationTime>
  </uws:jobref>
</uws:jobs>
"""


@pytest.mark.asyncio
async def test_job_list(
    client: AsyncClient,
    test_token: str,
    test_service: str,
    test_username: str,
    uws_factory: UWSFactory,
    mock_wobbly: MockWobbly,
) -> None:
    job_service = uws_factory.create_job_service()
    await job_service.create(test_token, SimpleParameters(name="Joe"))
    await job_service.create(
        test_token,
        SimpleParameters(name="Catherine"),
        run_id="some-run-id",
    )
    await job_service.create(test_token, SimpleParameters(name="Pat"))

    # Create an additional job for a different user, which shouldn't appear in
    # any of the lists.
    other_token = MockWobbly.make_token(test_service, "other-user")
    await job_service.create(other_token, SimpleParameters(name="Dominique"))

    # Adjust the creation time of the jobs so that searches are more
    # interesting.
    jobs = mock_wobbly.jobs[test_service][test_username]
    for i, job in enumerate(jobs.values()):
        hours = (2 - i) * 2
        job.creation_time = current_datetime() - timedelta(hours=hours)

    # Retrieve the job list and check it.
    r = await client.get("/test/jobs")
    assert r.status_code == 200
    assert r.headers["Content-Type"] == "application/xml"
    creation_times = [isodatetime(j.creation_time) for j in jobs.values()]
    creation_times.reverse()
    expected = FULL_JOB_LIST.strip().format(*creation_times)
    assert Jobs.from_xml(r.text) == Jobs.from_xml(expected)

    # Filter by recency.
    threshold = current_datetime() - timedelta(hours=1)
    r = await client.get(
        "/test/jobs", params={"after": isodatetime(threshold)}
    )
    assert r.status_code == 200
    assert r.headers["Content-Type"] == "application/xml"
    expected = RECENT_JOB_LIST.strip().format(creation_times[0])
    assert Jobs.from_xml(r.text) == Jobs.from_xml(expected)

    # Check case-insensitivity.
    result = r.text
    r = await client.get(
        "/test/jobs", params={"AFTER": isodatetime(threshold)}
    )
    assert r.text == result
    r = await client.get(
        "/test/jobs", params={"aFTer": isodatetime(threshold)}
    )
    assert r.text == result

    # Filter by count.
    r = await client.get("/test/jobs", params={"last": 1})
    assert r.status_code == 200
    assert r.headers["Content-Type"] == "application/xml"
    expected = RECENT_JOB_LIST.strip().format(creation_times[0])
    assert Jobs.from_xml(r.text) == Jobs.from_xml(expected)

    # Start the job.
    r = await client.post("/test/jobs/2/phase", data={"PHASE": "RUN"})
    assert r.status_code == 303
    assert r.headers["Location"] == "https://example.com/test/jobs/2"

    # Filter by phase.
    r = await client.get(
        "/test/jobs",
        params=[("PHASE", "EXECUTING"), ("PHASE", "QUEUED")],
    )
    assert r.status_code == 200
    assert r.headers["Content-Type"] == "application/xml"
    expected = QUEUED_JOB_LIST.strip().format(creation_times[1])
    assert Jobs.from_xml(r.text) == Jobs.from_xml(expected)



"""Test metrics tooling."""

from __future__ import annotations

import asyncio
import math
from datetime import UTC, datetime, timedelta
from enum import Enum
from typing import cast
from uuid import UUID

import pytest
from aiokafka import AIOKafkaConsumer, TopicPartition
from aiokafka.admin.client import AIOKafkaAdminClient, NewTopic
from aiokafka.errors import KafkaConnectionError
from dataclasses_avroschema.pydantic import AvroBaseModel
from pydantic import Field, HttpUrl
from schema_registry.client.client import AsyncSchemaRegistryClient
from schema_registry.serializers.message_serializer import (
    AsyncAvroMessageSerializer,
)
from structlog.testing import LogCapture
from time_machine import TimeMachineFixture

from safir.dependencies.metrics import EventDependency, EventMaker
from safir.kafka import (
    KafkaConnectionSettings,
    SchemaManagerSettings,
    SecurityProtocol,
)
from safir.metrics import (
    DisabledMetricsConfiguration,
    DuplicateEventError,
    EventManager,
    EventManagerUnintializedError,
    EventMetadata,
    EventPayload,
    EventPublisher,
    EventsConfiguration,
    KafkaEventManager,
    KafkaEventPublisher,
    KafkaMetricsConfiguration,
    KafkaTopicError,
)
from safir.metrics._exceptions import UnsupportedAvroSchemaError
from tests.support.kafka import KafkaClients, KafkaStack

from ..support.metrics import MetricsStack


class MyEvent(EventPayload):
    """An event payload."""

    foo: str
    duration: timedelta
    duration_union: timedelta | None


class Events(EventMaker):
    """A class to hold event publishers."""

    async def initialize(self, manager: EventManager) -> None:
        self.my_event = await manager.create_publisher("myevent", MyEvent)


async def subscribe_and_wait(consumer: AIOKafkaConsumer, pattern: str) -> None:
    """Avoid race condition when subscribing to topics.

    There is some race condition where an exception is raised if we start
    trying to consume messages too soon after we have subscribed to a topic
    or pattern, because partitions haven't been assigned to the consumer yet.
    """
    consumer.subscribe(pattern=pattern)
    await asyncio.sleep(0.5)


async def assert_from_kafka(
    consumer: AIOKafkaConsumer,
    schema_registry: AsyncSchemaRegistryClient,
    event: EventPublisher[MyEvent],
    foo: str,
    duration: timedelta,
) -> None:
    """Grab a message from kafka, deserialize it, and assert stuff about it."""
    message = await consumer.getone()
    assert isinstance(message.value, bytes)
    serializer = AsyncAvroMessageSerializer(schema_registry)
    deserialized_dict = await serializer.decode_message(message.value)
    assert deserialized_dict is not None
    event_class = cast("type[AvroBaseModel]", event._event_class)
    deserialized = event_class(**deserialized_dict)

    assert isinstance(deserialized, EventMetadata)
    assert isinstance(deserialized.id, UUID)
    assert deserialized.application == "testapp"

    # dataclasses-avroschema serializes python datetime's into avro
    # timestamp-millis's. The timestamp_ns and the timestamp fields could
    # differ by a millisecond due to differences between time.time_ns and the
    # passing of datetime objects through the dataclasses-avroschema
    # serialization and deserialization.
    deserialized_ms = math.trunc(deserialized.timestamp_ns / 1e6)
    timestamp_ms = deserialized.timestamp.timestamp() * 1e3
    difference = abs(deserialized_ms - timestamp_ms)
    assert difference <= 1

    # Mypy can't deal with the fact that deserialized has both EventMetadata
    # AND MyEvent as bases, so let's pretend we don't know either :(
    # https://github.com/python/mypy/issues/12314
    assert getattr(deserialized, "foo", None) == foo
    assert getattr(deserialized, "duration", None) == duration


async def integration_test(
    manager: EventManager,
    schema_registry_client: AsyncSchemaRegistryClient,
    kafka_consumer: AIOKafkaConsumer,
    kafka_admin_client: AIOKafkaAdminClient,
    *,
    create_topic: bool = True,
) -> None:
    """Test the interaction of the EventManager and the EventsDependency."""
    if create_topic:
        topic = NewTopic(
            name="what.ever.testapp",
            num_partitions=1,
            replication_factor=1,
        )
        await kafka_admin_client.create_topics([topic])

    # Initialize the event manager
    await manager.initialize()

    # Create an events dependency and initialize it with the transport config
    events_dependency = EventDependency(Events())
    await events_dependency.initialize(manager=manager)

    # Publish events
    event = events_dependency.events.my_event
    await event.publish(
        MyEvent(
            foo="bar1",
            duration=timedelta(seconds=2, milliseconds=123),
            duration_union=None,
        )
    )
    published = await event.publish(
        MyEvent(
            foo="bar2",
            duration=timedelta(seconds=2, milliseconds=456),
            duration_union=None,
        )
    )
    schema = published.avro_schema_to_python()
    assert schema["name"] == "myevent"
    assert schema["namespace"] == "what.ever.testapp"

    # Set up a kafka consumer
    expected_topic = "what.ever.testapp"
    await subscribe_and_wait(consumer=kafka_consumer, pattern=expected_topic)
    await kafka_consumer.seek_to_beginning()

    # Assert stuff
    assert isinstance(event, KafkaEventPublisher)
    assert event._publisher.topic == expected_topic
    await assert_from_kafka(
        kafka_consumer,
        schema_registry_client,
        event,
        "bar1",
        timedelta(seconds=2.123),
    )
    await assert_from_kafka(
        kafka_consumer,
        schema_registry_client,
        event,
        "bar2",
        timedelta(seconds=2.456),
    )

    await manager.aclose()


@pytest.mark.asyncio
async def test_managed_storage(
    kafka_stack: KafkaStack,
    kafka_clients: KafkaClients,
) -> None:
    """Publish events to actual storage and read them back and verify them."""
    config = KafkaMetricsConfiguration(
        application="testapp",
        events=EventsConfiguration(topic_prefix="what.ever"),
        kafka=kafka_stack.kafka_connection_settings,
        schema_manager=kafka_stack.schema_manager_settings,
    )

    # Construct an event manager and intialize our events dependency
    manager = config.make_manager()
    await integration_test(
        manager,
        kafka_clients.schema_registry_client,
        kafka_clients.kafka_consumer,
        kafka_clients.kafka_admin_client,
    )

    # Make sure storage is cleaned up
    assert manager._admin_client._closed
    assert not await manager._broker.ping(timeout=1)


@pytest.mark.asyncio
async def test_unmanaged_storage(
    kafka_clients: KafkaClients,
) -> None:
    """Publish events to actual storage and read them back and verify them."""
    manager = KafkaEventManager(
        application="testapp",
        topic_prefix="what.ever",
        kafka_broker=kafka_clients.kafka_broker,
        kafka_admin_client=kafka_clients.kafka_admin_client,
        schema_manager=kafka_clients.schema_manager,
        manage_kafka_broker=False,
    )
    await integration_test(
        manager,
        kafka_clients.schema_registry_client,
        kafka_clients.kafka_consumer,
        kafka_clients.kafka_admin_client,
    )

    # Make sure storage is NOT cleaned up
    assert await manager._broker.ping(timeout=1)


@pytest.mark.asyncio
async def test_topic_not_created(
    kafka_clients: KafkaClients,
) -> None:
    kafka_broker = kafka_clients.kafka_broker
    kafka_consumer = kafka_clients.kafka_consumer
    kafka_admin_client = kafka_clients.kafka_admin_client
    schema_registry_client = kafka_clients.schema_registry_client
    schema_manager = kafka_clients.schema_manager

    manager = KafkaEventManager(
        application="testapp",
        topic_prefix="what.ever",
        kafka_broker=kafka_broker,
        kafka_admin_client=kafka_admin_client,
        schema_manager=schema_manager,
        manage_kafka_broker=False,
    )

    with pytest.raises(KafkaTopicError):
        await integration_test(
            manager,
            schema_registry_client,
            kafka_consumer,
            kafka_admin_client,
            create_topic=False,
        )


@pytest.mark.asyncio
async def test_create_before_initialize(
    kafka_clients: KafkaClients,
) -> None:
    kafka_admin_client = kafka_clients.kafka_admin_client
    kafka_broker = kafka_clients.kafka_broker
    schema_manager = kafka_clients.schema_manager

    manager = KafkaEventManager(
        application="testapp",
        topic_prefix="what.ever",
        kafka_broker=kafka_broker,
        kafka_admin_client=kafka_admin_client,
        schema_manager=schema_manager,
        manage_kafka_broker=False,
    )

    with pytest.raises(EventManagerUnintializedError):
        await manager.create_publisher("myevent", MyEvent)


@pytest.mark.asyncio
async def test_duplicate_event(event_manager: EventManager) -> None:
    class MyOtherEvent(EventPayload):
        blah: int

    await event_manager.create_publisher("myevent", MyEvent)
    with pytest.raises(DuplicateEventError):
        await event_manager.create_publisher("myevent", MyOtherEvent)


@pytest.mark.asyncio
async def test_invalid_payload(event_manager: EventManager) -> None:
    class MyEnum(Enum):
        case1 = "case1"
        case2 = "case2"

    class MyInvalidEvent(EventPayload):
        good_simple_field: str = Field()
        good_union_field: str | None = Field()
        good_enum_field: MyEnum = Field()
        bad_list_field: list[str] = Field()
        bad_union_field: dict[str, str] | None = Field()
        bad_dict_field: dict[str, str] = Field()

    with pytest.raises(UnsupportedAvroSchemaError) as excinfo:
        await event_manager.create_publisher("myinvalidevent", MyInvalidEvent)
    err = str(excinfo.value)

    assert "bad_list_field" in err
    assert "bad_union_field" in err
    assert "bad_dict_field" in err

    assert "good_simple_field" not in err
    assert "good_union_field" not in err
    assert "good_enum_field" not in err


@pytest.mark.asyncio
async def test_disable() -> None:
    config = DisabledMetricsConfiguration(
        enabled=False,
        application="testapp",
        events=EventsConfiguration(topic_prefix="what.ever"),
    )
    manager = config.make_manager()

    await manager.initialize()
    event = await manager.create_publisher("myevent", MyEvent)

    await event.publish(
        MyEvent(foo="bar1", duration=timedelta(seconds=1), duration_union=None)
    )
    await event.publish(
        MyEvent(foo="bar2", duration=timedelta(seconds=1), duration_union=None)
    )
    await manager.aclose()


@pytest.mark.asyncio
async def test_bad_kafka_from_start_no_raises(
    log_output: LogCapture,
    time_machine: TimeMachineFixture,
) -> None:
    # This has to be long enough so that we won't exhaust it naturally during
    # the execution of the test.
    backoff_interval = timedelta(hours=1)

    bad_kafka_settings = KafkaConnectionSettings(
        bootstrap_servers="nope:1234,nada:5678",
        security_protocol=SecurityProtocol.PLAINTEXT,
    )

    bad_schema_manager_settings = SchemaManagerSettings(
        registry_url=HttpUrl("https://nope.com")
    )
    config = KafkaMetricsConfiguration(
        application="testapp",
        enabled=True,
        events=EventsConfiguration(topic_prefix="what.ever"),
        kafka=bad_kafka_settings,
        schema_manager=bad_schema_manager_settings,
        backoff_interval=backoff_interval,
    )

    manager = config.make_manager()

    await manager.initialize()

    # Create an events dependency
    events_dependency = EventDependency(Events())
    await events_dependency.initialize(manager=manager)

    # Publish events, nothing should blow up
    event = events_dependency.events.my_event
    await event.publish(
        MyEvent(
            foo="bar1",
            duration=timedelta(seconds=2, milliseconds=123),
            duration_union=None,
        )
    )
    await event.publish(
        MyEvent(
            foo="bar2",
            duration=timedelta(seconds=2, milliseconds=456),
            duration_union=None,
        )
    )

    await manager.aclose()

    # We should have logged one publish error
    error_logs = [
        entry
        for entry in log_output.entries
        if entry.get("attempted_operation") == "publish"
    ]

    # Move time to after the backoff window. This isn't EXACTLY when the
    # backoff interval expires, but it's close enough for testing.
    time_machine.move_to(datetime.now(tz=UTC) + backoff_interval, tick=True)

    await event.publish(
        MyEvent(
            foo="bar3",
            duration=timedelta(seconds=2, milliseconds=123),
            duration_union=None,
        )
    )
    await event.publish(
        MyEvent(
            foo="bar4",
            duration=timedelta(seconds=2, milliseconds=456),
            duration_union=None,
        )
    )

    # We should only have one more error log
    error_logs = [
        entry
        for entry in log_output.entries
        if entry.get("attempted_operation") == "publish"
    ]
    assert len(error_logs) == 2


@pytest.mark.asyncio
async def test_bad_kafka_from_start_raises() -> None:
    bad_kafka_settings = KafkaConnectionSettings(
        bootstrap_servers="nope:1234,nada:5678",
        security_protocol=SecurityProtocol.PLAINTEXT,
    )

    bad_schema_manager_settings = SchemaManagerSettings(
        registry_url=HttpUrl("https://nope.com")
    )
    config = KafkaMetricsConfiguration(
        application="testapp",
        enabled=True,
        events=EventsConfiguration(topic_prefix="what.ever"),
        kafka=bad_kafka_settings,
        schema_manager=bad_schema_manager_settings,
        raise_on_error=True,
    )

    manager = config.make_manager()

    with pytest.raises(KafkaConnectionError):
        await manager.initialize()


@pytest.mark.asyncio
async def test_bad_kafka_after_initialize(
    resiliency_metrics_stack: MetricsStack,
    log_output: LogCapture,
    time_machine: TimeMachineFixture,
) -> None:
    event_manager = resiliency_metrics_stack.event_manager
    kafka_container = resiliency_metrics_stack.kafka_stack.kafka_container
    schema_registry_client = (
        resiliency_metrics_stack.kafka_clients.schema_registry_client
    )
    max_batch_size = resiliency_metrics_stack.kafka_max_batch_size
    backoff_interval = resiliency_metrics_stack.metrics_config.backoff_interval
    kafka_connection_settings = (
        resiliency_metrics_stack.kafka_stack.kafka_connection_settings
    )

    # We're gonna call a method on this, so tell the type checker it's not None
    assert event_manager._broker._producer

    # Create an events dependency
    events_dependency = EventDependency(Events())
    await events_dependency.initialize(manager=event_manager)

    event = events_dependency.events.my_event

    # This should make it to kafka
    await event.publish(
        MyEvent(
            foo="working",
            duration=timedelta(seconds=2, milliseconds=123),
            duration_union=None,
        )
    )
    await event_manager._broker._producer.flush()

    # Stop Kafka
    kafka_container.stop_container()

    # Quickly sending a bunch more messages than max_batch_size is set to
    # should get us an exception here. None of these should make it to Kafka.
    #
    # Sometimes, exceptions from aiokafka will be raised here, and sometimes
    # they will be raised in futures. If we send so many messages so quickly
    # that the message batch gets filled up, then send calls will block until
    # the batch is sent, and we'll get an exception here if the batch can't be
    # sent. If a batch never gets filled up before getting sent, then the
    # exception will be sitting in a future.
    for _ in range(max_batch_size + 5):
        await event.publish(
            MyEvent(
                foo="stopped",
                duration=timedelta(seconds=2, milliseconds=456),
                duration_union=None,
            )
        )
    await event_manager._broker._producer.flush()

    # This shouldn't raise an exception, but it shouldn't make it to kafka.
    await event.publish(
        MyEvent(
            foo="stopped",
            duration=timedelta(seconds=2, milliseconds=456),
            duration_union=None,
        )
    )
    await event_manager._broker._producer.flush()

    # There should be one error event in the logs
    error_logs = [
        entry
        for entry in log_output.entries
        if entry["log_level"] == "error"
        and entry["attempted_operation"] == "publish"
    ]
    assert len(error_logs) == 1

    # Move time to after the backoff window. This isn't EXACTLY when the
    # backoff interval expires, but it's close enough for testing.
    time_machine.move_to(datetime.now(tz=UTC) + backoff_interval, tick=True)

    # These should also not make it to Kafka, but there should be one more
    # error log because we're past the backoff window
    for _ in range(max_batch_size + 5):
        await event.publish(
            MyEvent(
                foo="in_backoff",
                duration=timedelta(seconds=2, milliseconds=456),
                duration_union=None,
            )
        )
    await event_manager._broker._producer.flush()

    backoff_down_error_logs = [
        entry
        for entry in log_output.entries
        if entry["log_level"] == "error"
        and entry["attempted_operation"] == "publish"
    ]
    assert len(backoff_down_error_logs) == len(error_logs) + 1

    # Start kafka again
    kafka_container.start_container_again()

    # These should also not make it to Kafka, even though it's back up, because
    # we're still in the backoff window
    for _ in range(max_batch_size + 5):
        await event.publish(
            MyEvent(
                foo="in_backoff_again",
                duration=timedelta(seconds=2, milliseconds=456),
                duration_union=None,
            )
        )

    # We still should not have any more error logs because we're still in the
    # backoff window
    backoff_up_error_logs = [
        entry
        for entry in log_output.entries
        if entry["log_level"] == "error"
        and entry["attempted_operation"] == "publish"
    ]
    await event_manager._broker._producer.flush()
    assert len(backoff_up_error_logs) == len(backoff_down_error_logs)

    # Move time to after the next backoff window. This isn't EXACTLY when the
    # backoff interval expires, but it's close enough for testing.
    time_machine.move_to(datetime.now(tz=UTC) + backoff_interval, tick=True)

    # These should all make it to Kafka
    for _ in range(max_batch_size + 5):
        await event.publish(
            MyEvent(
                foo="started_again",
                duration=timedelta(seconds=2, milliseconds=456),
                duration_union=None,
            )
        )
    await event_manager._broker._producer.flush()

    # There shouldn't be any more error logs
    new_error_logs = [
        entry for entry in log_output.entries if entry["log_level"] == "error"
    ]
    assert len(new_error_logs) == len(backoff_up_error_logs)

    consumer = AIOKafkaConsumer(
        **kafka_connection_settings.to_aiokafka_params(),
        client_id="pytest-consumer",
    )

    # Verify messages made it to Kafka
    try:
        await consumer.start()
        await subscribe_and_wait(consumer, "what.ever.testapp")
        await consumer.seek_to_beginning()
        topic = TopicPartition(topic="what.ever.testapp", partition=0)
        message_groups = await consumer.getmany(
            topic,
            timeout_ms=10000,
        )
    finally:
        await consumer.stop()

    messages = message_groups[topic]
    serializer = AsyncAvroMessageSerializer(schema_registry_client)
    deserialized = [
        await serializer.decode_message(message.value) for message in messages
    ]

    working = [m for m in deserialized if m and m.get("foo") == "working"]
    started_again = [
        m for m in deserialized if m and m.get("foo") == "started_again"
    ]

    # Kafka should have the only the messages from before we stopped the
    # container, and the messages from after we restarted the container and
    # after the backoff interval.
    assert len(deserialized) == max_batch_size + 5 + 1

    assert len(working) == 1
    assert len(started_again) == max_batch_size + 5



"""Pydantic models for GitHub webhook payloads."""

from __future__ import annotations

from enum import StrEnum

from pydantic import BaseModel, Field

from .models import (
    GitHubCheckRunModel,
    GitHubCheckSuiteModel,
    GitHubPullRequestModel,
    GitHubRepositoryModel,
)

__all__ = [
    "GitHubAppInstallationEventAction",
    "GitHubAppInstallationEventModel",
    "GitHubAppInstallationEventRepoModel",
    "GitHubAppInstallationModel",
    "GitHubAppInstallationRepositoriesEventAction",
    "GitHubAppInstallationRepositoriesEventModel",
    "GitHubCheckRunEventAction",
    "GitHubCheckRunEventModel",
    "GitHubCheckSuiteEventAction",
    "GitHubCheckSuiteEventModel",
    "GitHubPullRequestEventAction",
    "GitHubPullRequestEventModel",
    "GitHubPushEventModel",
]


class GitHubAppInstallationModel(BaseModel):
    """A Pydantic model for the ``installation`` field found in webhook
    payloads for GitHub Apps.
    """

    id: int = Field(description="The installation ID.")


class GitHubPushEventModel(BaseModel):
    """A Pydantic model for the ``push`` event webhook when a commit or
    tag is pushed.

    https://docs.github.com/en/webhooks/webhook-events-and-payloads#push
    """

    repository: GitHubRepositoryModel = Field(
        description="The repository that was pushed to."
    )

    installation: GitHubAppInstallationModel = Field(
        description="Information about the GitHub App installation."
    )

    ref: str = Field(
        description=(
            "The full git ref that was pushed. Example: refs/heads/main or "
            "refs/tags/v3.14.1."
        ),
        examples=["refs/heads/main"],
    )

    before: str = Field(
        description="The SHA of the most recent commit on ref before the push."
    )

    after: str = Field(
        description="The SHA of the most recent commit on ref after the push."
    )


class GitHubAppInstallationEventRepoModel(BaseModel):
    """A pydantic model for repository objects used by
    `GitHubAppInstallationRepositoriesEventModel`.

    https://docs.github.com/en/webhooks/webhook-events-and-payloads#installation
    """

    name: str = Field(
        description="The name of the repository, e.g. 'times-square'."
    )

    full_name: str = Field(
        description=(
            "The full name of the repository, e.g. 'lsst-sqre/times-square'."
        )
    )

    @property
    def owner_name(self) -> str:
        """The name of the repository owner."""
        return self.full_name.split("/")[0]


class GitHubAppInstallationEventAction(StrEnum):
    """The action performed on an GitHub App ``installation`` webhook
    (`GitHubAppInstallationEventModel`).
    """

    created = "created"
    """Someone installed a GitHub App on a user or organization account."""

    deleted = "deleted"
    """Someone uninstalled a GitHub App on a user or organization account."""

    new_permissions_accepted = "new_permissions_accepted"
    """Someone granted new permissions to a GitHub App."""

    suspend = "suspend"
    """Someone blocked access by a GitHub App to their user or org account."""

    unsuspend = "unsuspend"
    """Someone unblocked access by a GitHub App to their user or org
    account.
    """


class GitHubAppInstallationEventModel(BaseModel):
    """A Pydantic model for an ``installation`` webhook.

    https://docs.github.com/en/webhooks/webhook-events-and-payloads#installation
    """

    action: GitHubAppInstallationEventAction = Field(
        description="Action performed."
    )

    repositories: list[GitHubAppInstallationEventRepoModel] = Field(
        description="Repositories accessible to this installation."
    )

    installation: GitHubAppInstallationModel = Field(
        description="Information about the GitHub App installation."
    )


class GitHubAppInstallationRepositoriesEventAction(StrEnum):
    """The action performed on a GitHub App ``installation_repositories``
    webhook (`GitHubAppInstallationRepositoriesEventModel`).
    """

    #: Someone added a repository to an installation.
    added = "added"

    #: Someone removed a repository from an installation.
    removed = "removed"


class GitHubAppInstallationRepositoriesEventModel(BaseModel):
    """A Pydantic model for a ``installation_repositories`` webhook.

    https://docs.github.com/en/webhooks/webhook-events-and-payloads#installation_repositories
    """

    action: GitHubAppInstallationRepositoriesEventAction = Field(
        description="Action performed on the installation."
    )

    repositories_added: list[GitHubAppInstallationEventRepoModel] = Field(
        description="Repositories added to the installation."
    )

    repositories_removed: list[GitHubAppInstallationEventRepoModel] = Field(
        description="Repositories removed from the installation."
    )

    installation: GitHubAppInstallationModel = Field(
        description="Information about the GitHub App installation."
    )


class GitHubPullRequestEventAction(StrEnum):
    """The action performed on a GitHub ``pull_request`` webhook
    (`GitHubPullRequestEventModel`).
    """

    assigned = "assigned"
    """A pull request was assigned to a user."""

    auto_merge_disabled = "auto_merge_disabled"
    """Auto merge was disabled for a pull request."""

    auto_merge_enabled = "auto_merge_enabled"
    """Auto merge was enabled for a pull request."""

    closed = "closed"
    """A pull request was closed."""

    converted_to_draft = "converted_to_draft"
    """A pull request was converted to a draft."""

    demilestoned = "demilestoned"
    """A pull request was removed from a milestone."""

    dequeued = "dequeued"
    """A pull request was removed from the merge queue."""

    edited = "edited"
    """The title or body of a pull request was edited."""

    labeled = "labeled"
    """A label was added to a pull request."""

    locked = "locked"
    """Conversation on a pull request was locked."""

    milestoned = "milestoned"
    """A pull request was added to a milestone."""

    opened = "opened"
    """A pull request was created."""

    ready_for_review = "ready_for_review"
    """A draft pull request was marked as ready for review."""

    reopened = "reopened"
    """A pull request was reopened."""

    review_request_removed = "review_request_removed"
    """A request for review by a person or team was removed from a pull
    request.
    """

    review_requested = "review_requested"
    """Review by a person or team was requested for a pull request."""

    synchronize = "synchronize"
    """A pull request's head branch was updated. For example, the head branch
    was updated from the base branch or new commits were pushed to the head
    branch.
    """

    unassigned = "unassigned"
    """A user was unassigned from a pull request."""

    unlabeled = "unlabeled"
    """A label was removed from a pull request."""

    unlocked = "unlocked"
    """Conversation on a pull request was unlocked."""


class GitHubPullRequestEventModel(BaseModel):
    """A Pydantic model for a ``pull_request`` webhook.

    https://docs.github.com/en/webhooks/webhook-events-and-payloads#pull_request
    """

    repository: GitHubRepositoryModel = Field(
        description="The repository that the pull request was opened against."
    )

    installation: GitHubAppInstallationModel = Field(
        description="Information about the GitHub App installation."
    )

    action: GitHubPullRequestEventAction = Field(
        description="The action that was performed.",
    )

    number: int = Field(description="Pull request number")

    pull_request: GitHubPullRequestModel = Field(
        description="Information about the pull request."
    )


class GitHubCheckSuiteEventAction(StrEnum):
    """The action performed in a GitHub ``check_suite`` webhook
    (`GitHubCheckSuiteEventModel`).
    """

    completed = "completed"
    """All check runs in a check suite have completed, and a conclusion is
    available.
    """

    requested = "requested"
    """Someone requested to run a check suite."""

    rerequested = "rerequested"
    """Someone requested to re-run the check runs in a check suite."""


class GitHubCheckSuiteEventModel(BaseModel):
    """A Pydantic model for the ``check_suite`` webhook payload.

    https://docs.github.com/en/webhooks/webhook-events-and-payloads#check_suite
    """

    action: GitHubCheckSuiteEventAction = Field(
        description="The action performed.",
    )

    check_suite: GitHubCheckSuiteModel = Field(
        description="Information about the check suite."
    )

    repository: GitHubRepositoryModel = Field(
        description="The repository that the check suite was run against."
    )

    installation: GitHubAppInstallationModel = Field(
        description="Information about the GitHub App installation."
    )


class GitHubCheckRunEventAction(StrEnum):
    """The action performed in a GitHub ``check_run`` webhook
    (`GitHubCheckRunEventModel`).
    """

    completed = "completed"
    """A check run was completed and a conclusion is available."""

    created = "created"
    """A new check run was created."""

    requested_action = "requested_action"
    """A check run completed, and someone requested a followup action that
    your app provides.
    """

    rerequested = "rerequested"
    """Someone requested to re-run a check run."""


class GitHubCheckRunEventModel(BaseModel):
    """A Pydantic model for the ``check_run`` webhook payload.

    https://docs.github.com/en/webhooks/webhook-events-and-payloads#check_run
    """

    action: GitHubCheckRunEventAction = Field(
        description="The action that was performed.",
    )

    repository: GitHubRepositoryModel = Field(
        description="The repository that the check run was run against."
    )

    installation: GitHubAppInstallationModel = Field(
        description="Information about the GitHub App installation."
    )

    check_run: GitHubCheckRunModel = Field(
        description="Information about the check run."
    )



"""Tests for mock publishers."""

from typing import cast

import pytest
from pydantic import ConfigDict

from safir.metrics import (
    ANY,
    NOT_NONE,
    EventPayload,
    EventsConfiguration,
    MockEventPublisher,
    MockMetricsConfiguration,
    NotPublishedConsecutivelyError,
    NotPublishedError,
    PublishedCountError,
    PublishedList,
)


async def publish() -> PublishedList:
    """Mock publish some events and return the PublishedList."""
    config = MockMetricsConfiguration(
        enabled=False,
        mock=True,
        application="testapp",
        events=EventsConfiguration(topic_prefix="what.ever"),
    )
    manager = config.make_manager()

    class SomeEvent(EventPayload):
        model_config = ConfigDict(ser_json_timedelta="float")

        foo: str
        count: int
        duration: float | None

    await manager.initialize()
    pub = await manager.create_publisher("someevent", SomeEvent)

    await pub.publish(SomeEvent(foo="foo1", count=1, duration=1.234))
    await pub.publish(SomeEvent(foo="foo2", count=2, duration=2.345))
    await pub.publish(SomeEvent(foo="foo3", count=3, duration=3.456))
    await pub.publish(SomeEvent(foo="foo4", count=4, duration=None))
    await pub.publish(SomeEvent(foo="foo5", count=5, duration=5.678))

    await manager.aclose()

    return cast("MockEventPublisher", pub).published


@pytest.mark.asyncio
async def test_assert_one() -> None:
    pub = await publish()
    pub.assert_published(
        [
            {"foo": "foo1", "count": 1, "duration": 1.234},
        ]
    )
    pub.assert_published(
        [
            {"foo": "foo1", "count": 1, "duration": 1.234},
        ],
        any_order=False,
    )

    with pytest.raises(NotPublishedError):
        pub.assert_published(
            [
                {"foo": "foo1", "count": 1, "duration": 1},
            ]
        )


@pytest.mark.asyncio
async def test_assert_unordered() -> None:
    pub = await publish()
    pub.assert_published(
        [
            {"foo": "foo2", "count": 2, "duration": 2.345},
            {"foo": "foo1", "count": 1, "duration": 1.234},
            {"foo": "foo3", "count": 3, "duration": 3.456},
        ],
        any_order=True,
    )

    with pytest.raises(NotPublishedConsecutivelyError):
        pub.assert_published(
            [
                {"foo": "foo2", "count": 2, "duration": 2.345},
                {"foo": "foo1", "count": 1, "duration": 1.234},
                {"foo": "foo3", "count": 3, "duration": 3.456},
            ],
        )


@pytest.mark.asyncio
async def test_only_counted_once() -> None:
    pub = await publish()

    with pytest.raises(NotPublishedError):
        pub.assert_published(
            [
                {"foo": "foo1", "count": 1, "duration": 1.234},
                {"foo": "foo1", "count": 1, "duration": 1.234},
            ],
            any_order=True,
        )


@pytest.mark.asyncio
async def test_assert_ordered() -> None:
    pub = await publish()
    pub.assert_published(
        [
            {"foo": "foo1", "count": 1, "duration": 1.234},
            {"foo": "foo2", "count": 2, "duration": 2.345},
            {"foo": "foo3", "count": 3, "duration": 3.456},
        ]
    )

    with pytest.raises(NotPublishedConsecutivelyError):
        pub.assert_published(
            [
                {"foo": "foo1", "count": 1, "duration": 1.234},
                {"foo": "foo3", "count": 3, "duration": 3.456},
            ],
        )


@pytest.mark.asyncio
async def test_assert_extra() -> None:
    pub = await publish()
    with pytest.raises(NotPublishedError):
        pub.assert_published(
            [
                {"foo": "foo1", "count": 1, "duration": 1.234},
                {"foo": "foo2", "count": 2, "duration": 2.345},
                {"foo": "foo3", "count": 3, "duration": 3.456},
                {"foo": "nope", "count": 3, "duration": 3.456},
            ]
        )
    with pytest.raises(NotPublishedError):
        pub.assert_published(
            [
                {"foo": "foo1", "count": 1, "duration": 1.234},
                {"foo": "foo2", "count": 2, "duration": 2.345},
                {"foo": "foo3", "count": 3, "duration": 3.456},
                {"foo": "nope", "count": 3, "duration": 3.456},
            ],
            any_order=False,
        )


@pytest.mark.asyncio
async def test_assert_all() -> None:
    pub = await publish()
    pub.assert_published_all(
        [
            {"foo": "foo1", "count": 1, "duration": 1.234},
            {"foo": "foo2", "count": 2, "duration": 2.345},
            {"foo": "foo4", "count": 4, "duration": None},
            {"foo": "foo3", "count": 3, "duration": 3.456},
            {"foo": "foo5", "count": 5, "duration": 5.678},
        ],
        any_order=True,
    )

    pub.assert_published_all(
        [
            {"foo": "foo1", "count": 1, "duration": 1.234},
            {"foo": "foo2", "count": 2, "duration": 2.345},
            {"foo": "foo3", "count": 3, "duration": 3.456},
            {"foo": "foo4", "count": 4, "duration": None},
            {"foo": "foo5", "count": 5, "duration": 5.678},
        ],
    )

    with pytest.raises(PublishedCountError):
        pub.assert_published_all(
            [
                {"foo": "foo1", "count": 1, "duration": 1.234},
                {"foo": "foo2", "count": 2, "duration": 2.345},
                {"foo": "foo4", "count": 4, "duration": None},
                {"foo": "foo3", "count": 3, "duration": 3.456},
            ]
        )

    with pytest.raises(NotPublishedConsecutivelyError):
        pub.assert_published_all(
            [
                {"foo": "foo1", "count": 1, "duration": 1.234},
                {"foo": "foo2", "count": 2, "duration": 2.345},
                {"foo": "foo4", "count": 4, "duration": None},
                {"foo": "foo3", "count": 3, "duration": 3.456},
                {"foo": "foo5", "count": 5, "duration": 5.678},
            ],
        )


@pytest.mark.asyncio
async def test_any_and_not_none() -> None:
    pub = await publish()
    pub.assert_published(
        [
            {"foo": "foo1", "count": 1, "duration": 1.234},
            {"foo": "foo2", "count": 2, "duration": ANY},
            {"foo": "foo3", "count": 3, "duration": 3.456},
            {"foo": "foo4", "count": 4, "duration": ANY},
        ]
    )

    pub.assert_published(
        [
            {"foo": "foo1", "count": 1, "duration": 1.234},
            {"foo": "foo2", "count": 2, "duration": NOT_NONE},
            {"foo": "foo3", "count": 3, "duration": 3.456},
            {"foo": "foo4", "count": 4, "duration": ANY},
        ]
    )

    with pytest.raises(NotPublishedError):
        pub.assert_published(
            [
                {"foo": "foo1", "count": 1, "duration": 1.234},
                {"foo": "foo2", "count": 2, "duration": NOT_NONE},
                {"foo": "foo3", "count": 3, "duration": 3.456},
                {"foo": "foo4", "count": 4, "duration": NOT_NONE},
            ]
        )



"""Utilities for configuring structlog-based logging."""

from __future__ import annotations

import logging
import logging.config
import sys
from typing import Any

import structlog
from structlog.stdlib import add_log_level
from structlog.types import EventDict

try:
    from safir.dependencies import logger as logger_dependency

    _UPDATE_DEPENDENCY = True
except Exception:
    _UPDATE_DEPENDENCY = False

from ._models import LogLevel, Profile

__all__ = [
    "add_log_severity",
    "configure_logging",
]


def add_log_severity(
    logger: logging.Logger, method_name: str, event_dict: EventDict
) -> EventDict:
    """Add the log level to the event dict as ``severity``.

    Intended for use as a structlog processor.

    This is the same as `structlog.stdlib.add_log_level` except that it
    uses the ``severity`` key rather than ``level`` for compatibility with
    Google Log Explorer and its automatic processing of structured logs.

    Parameters
    ----------
    logger
        The wrapped logger object.
    method_name
        The name of the wrapped method (``warning`` or ``error``, for
        example).
    event_dict
        Current context and current event. This parameter is also modified in
        place, matching the normal behavior of structlog processors.

    Returns
    -------
    ``structlog.types.EventDict``
        The modified ``structlog.types.EventDict`` with the added key.
    """
    severity = add_log_level(logger, method_name, {})["level"]
    event_dict["severity"] = severity
    return event_dict


def configure_logging(
    *,
    name: str,
    profile: Profile | str = Profile.production,
    log_level: LogLevel | str = LogLevel.INFO,
    add_timestamp: bool = False,
) -> None:
    """Configure logging and structlog.

    Parameters
    ----------
    name
        Name of the logger, which is typically the name of your application's
        root namespace.
    profile
        The name of the application profile:

        development
            Log messages are formatted for easier reading on the terminal.
        production
            Log messages are formatted as JSON objects.

        May be given as a `Profile` enum value (preferred) or a string.
    log_level
        The Python log level. May be given as a `LogLevel` enum (preferred)
        or a case-insensitive string.
    add_timestamp
        Whether to add an ISO-format timestamp to each log message.

    Notes
    -----
    This function helps you configure a useful logging set up for your
    application that's based on structlog_.

    First, it configures the logger for your application to log to STDOUT.
    Second, it configures the formatting of your log messages through
    structlog.

    In development mode, messages are key-value formatted, like this:

    .. code-block:: text

       [info     ] Hello world                    [myapp] answer=42

    Here, "Hello world" is the message. ``answer=42`` is a value bound to the
    logger.

    In production mode, messages are formatted as JSON objects:

    .. code-block:: text

       {"answer": 42, "event": "Hello world", "logger": "myapp",
       "severity": "info"}

    Examples
    --------
    .. code-block:: python

       import structlog
       from safir.logging import configure_logging


       configure_logging(name="mybot")
       logger = structlog.get_logger("mybot")
       logger.info("Hello world")
    """
    if not isinstance(log_level, LogLevel):
        log_level = LogLevel[log_level.upper()]
    if not isinstance(profile, Profile):
        profile = Profile[profile]

    stream_handler = logging.StreamHandler(stream=sys.stdout)
    stream_handler.setFormatter(logging.Formatter("%(message)s"))
    logger = logging.getLogger(name)
    logger.handlers = []
    logger.addHandler(stream_handler)
    logger.setLevel(log_level.value)

    processors: list[Any] = [
        structlog.stdlib.filter_by_level,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.PositionalArgumentsFormatter(),
    ]
    if add_timestamp:
        processors.append(structlog.processors.TimeStamper(fmt="iso"))
    processors.extend(
        [
            structlog.processors.StackInfoRenderer(),
            structlog.processors.UnicodeDecoder(),
        ]
    )
    if profile == Profile.production:
        # JSON-formatted logging
        processors.append(add_log_severity)
        processors.append(structlog.processors.format_exc_info)
        processors.append(structlog.processors.JSONRenderer())
    else:
        # Key-value formatted logging
        processors.append(structlog.stdlib.add_log_level)
        processors.append(structlog.dev.ConsoleRenderer())

    structlog.configure(
        processors=processors,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        cache_logger_on_first_use=True,
    )

    # Set the configured name for the global logger.
    if _UPDATE_DEPENDENCY:
        logger_dependency._logger_name = name  # noqa: SLF001



"""Tests for the schema management functionality in the safir.kafka module."""

from typing import Any

import pytest
from dataclasses_avroschema.pydantic import AvroBaseModel
from pydantic import Field
from schema_registry.client import AsyncSchemaRegistryClient
from schema_registry.serializers.message_serializer import (
    AsyncAvroMessageSerializer,
)

from safir.kafka import (
    Compatibility,
    IncompatibleSchemaError,
    InvalidAvroNameError,
    PydanticSchemaManager,
    SchemaInfo,
    UnknownSchemaError,
)
from safir.kafka._schema_registry_config import SchemaManagerSettings
from tests.support.kafka import KafkaClients, KafkaStack


async def assert_round_trip(
    manager: PydanticSchemaManager,
    model: type[AvroBaseModel],
    settings: SchemaManagerSettings,
    **kwargs: Any,
) -> SchemaInfo:
    info = await manager.register_model(model)
    original = model(**kwargs)
    serialized = await manager.serialize(original)

    # deserialize by getting the schema from the registry
    fresh_registry_client = AsyncSchemaRegistryClient(
        **settings.to_registry_params()
    )
    fresh_serializer = AsyncAvroMessageSerializer(fresh_registry_client)
    raw = await fresh_serializer.decode_message(serialized)
    assert raw is not None
    deserialized = model(**raw)
    assert deserialized == original
    return info


@pytest.mark.asyncio
async def test_no_metadata(
    kafka_stack: KafkaStack,
    kafka_clients: KafkaClients,
) -> None:
    class MyBareModel(AvroBaseModel):
        str_field: str
        int_field: int

    info = await assert_round_trip(
        kafka_clients.schema_manager,
        MyBareModel,
        kafka_stack.schema_manager_settings,
        str_field="blah",
        int_field=123,
    )
    assert info.subject == "MyBareModel"
    assert info.schema["name"] == "MyBareModel"
    assert info.schema.get("namespace", None) is None


@pytest.mark.asyncio
async def test_name_only_metadata(
    kafka_stack: KafkaStack,
    kafka_clients: KafkaClients,
) -> None:
    class MyModel(AvroBaseModel):
        str_field: str
        int_field: int

        class Meta:
            schema_name = "mymodelcustom"

    info = await assert_round_trip(
        kafka_clients.schema_manager,
        MyModel,
        kafka_stack.schema_manager_settings,
        str_field="blah",
        int_field=123,
    )
    assert info.subject == "mymodelcustom"
    assert info.schema["name"] == "mymodelcustom"
    assert info.schema.get("namespace", None) is None


@pytest.mark.asyncio
async def test_name_and_namespace_metadata(
    kafka_stack: KafkaStack,
    kafka_clients: KafkaClients,
) -> None:
    class MyOtherModel(AvroBaseModel):
        str_field: str
        int_field: int

        class Meta:
            schema_name = "myothermodelcustom"
            namespace = "my.other.namespace"

    info = await assert_round_trip(
        kafka_clients.schema_manager,
        MyOtherModel,
        kafka_stack.schema_manager_settings,
        str_field="blah",
        int_field=123,
    )
    assert info.subject == "my.other.namespace.myothermodelcustom"
    assert info.schema["name"] == "myothermodelcustom"
    assert info.schema["namespace"] == "my.other.namespace"


@pytest.mark.asyncio
async def test_namespace_only_metadata(
    kafka_stack: KafkaStack,
    kafka_clients: KafkaClients,
) -> None:
    class YetAnotherModel(AvroBaseModel):
        str_field: str
        int_field: int

        class Meta:
            namespace = "yetanothernamespace"

    info = await assert_round_trip(
        kafka_clients.schema_manager,
        YetAnotherModel,
        kafka_stack.schema_manager_settings,
        str_field="blah",
        int_field=123,
    )
    assert info.subject == "yetanothernamespace.YetAnotherModel"
    assert info.schema["name"] == "YetAnotherModel"
    assert info.schema["namespace"] == "yetanothernamespace"


@pytest.mark.asyncio
async def test_suffix(
    kafka_stack: KafkaStack,
) -> None:
    registry = AsyncSchemaRegistryClient(
        **kafka_stack.schema_manager_settings.to_registry_params()
    )
    schema_manager = PydanticSchemaManager(
        registry=registry,
        suffix="_test_suffix",
    )

    class MyModel(AvroBaseModel):
        str_field: str
        int_field: int

        class Meta:
            schema_name = "mymodel"
            namespace = "my.namespace"

    info = await assert_round_trip(
        schema_manager,
        MyModel,
        kafka_stack.schema_manager_settings,
        str_field="blah",
        int_field=123,
    )
    assert info.subject == "my.namespace.mymodel_test_suffix"
    assert info.schema["name"] == "mymodel"
    assert info.schema["namespace"] == "my.namespace"


@pytest.mark.asyncio
async def test_invalid_name(kafka_clients: KafkaClients) -> None:
    class YouGiveLove(AvroBaseModel):
        str_field: str
        int_field: int

        class Meta:
            schema_name = "a-bad-name"

    with pytest.raises(InvalidAvroNameError):
        await kafka_clients.schema_manager.register_model(YouGiveLove)


@pytest.mark.asyncio
async def test_forward_compatibility(kafka_clients: KafkaClients) -> None:
    class MyForwardModel(AvroBaseModel):
        field1: str
        field2: str

    await kafka_clients.schema_manager.register_model(
        MyForwardModel, compatibility=Compatibility.FORWARD
    )

    # Adding a field is forward compatible
    class MyForwardModel(AvroBaseModel):  # type: ignore[no-redef]
        field1: str
        field2: str
        field3: str

    await kafka_clients.schema_manager.register_model(MyForwardModel)

    # Adding another field is forward compatible
    class MyForwardModel(AvroBaseModel):  # type: ignore[no-redef]
        field1: str
        field2: str
        field3: str
        field4: str

    await kafka_clients.schema_manager.register_model(MyForwardModel)

    # Removing a field without a default is NOT foward compatible, even if
    # there is an identical grandparent or older schema.
    class MyForwardModel(AvroBaseModel):  # type: ignore[no-redef]
        field1: str
        field2: str
        field3: str

    with pytest.raises(
        IncompatibleSchemaError, match=r"READER_FIELD_MISSING_DEFAULT_VALUE"
    ):
        await kafka_clients.schema_manager.register_model(MyForwardModel)

    # Giving a field a default is forward compatible
    class MyForwardModel(AvroBaseModel):  # type: ignore[no-redef]
        field1: str
        field2: str
        field3: str
        field4: str = Field(default="blah")

    await kafka_clients.schema_manager.register_model(MyForwardModel)

    # Removing a field with a default is foward compatible
    class MyForwardModel(AvroBaseModel):  # type: ignore[no-redef]
        field1: str
        field2: str
        field3: str

    await kafka_clients.schema_manager.register_model(MyForwardModel)

    # Adding new fields is forward compatible
    class MyForwardModel(AvroBaseModel):  # type: ignore[no-redef]
        field1: str
        field2: str
        field3: str
        field4: str
        field5: str

    await kafka_clients.schema_manager.register_model(MyForwardModel)


@pytest.mark.asyncio
async def test_backward_compatibility(kafka_clients: KafkaClients) -> None:
    class MyBackwardModel(AvroBaseModel):
        field1: str
        field2: str

    await kafka_clients.schema_manager.register_model(
        MyBackwardModel, compatibility=Compatibility.BACKWARD
    )

    # Removing fields is backwards compatible
    class MyBackwardModel(AvroBaseModel):  # type: ignore[no-redef]
        field1: str

    await kafka_clients.schema_manager.register_model(MyBackwardModel)

    # Adding fields is NOT backwards compatible, even if an identical schema
    # has been previously registered
    class MyBackwardModel(AvroBaseModel):  # type: ignore[no-redef]
        field1: str
        field2: str

    with pytest.raises(
        IncompatibleSchemaError, match=r"READER_FIELD_MISSING_DEFAULT_VALUE"
    ):
        await kafka_clients.schema_manager.register_model(MyBackwardModel)

    # Adding a field with a default value is backwards compatible
    class MyBackwardModel(AvroBaseModel):  # type: ignore[no-redef]
        field1: str
        field2: str = Field(default="blah")

    await kafka_clients.schema_manager.register_model(MyBackwardModel)

    # Removing a default value is backwards compatible
    class MyBackwardModel(AvroBaseModel):  # type: ignore[no-redef]
        field1: str
        field2: str

    await kafka_clients.schema_manager.register_model(MyBackwardModel)


@pytest.mark.asyncio
async def test_unmanaged_error(kafka_clients: KafkaClients) -> None:
    class MyModel(AvroBaseModel):
        str_field: str
        int_field: int

        class Meta:
            namespace = "what.ever"

    original = MyModel(str_field="somestring", int_field=123)

    with pytest.raises(
        UnknownSchemaError,
        match=r"model: MyModel with subject: what.ever.MyModel",
    ):
        await kafka_clients.schema_manager.serialize(original)



from documenteer.conf.guide import *

# Disable JSON schema because it doesn't seem that useful and apparently can't
# deal with generics, so it produces warnings for the UWS Job model.
autodoc_pydantic_model_show_json = False

html_sidebars["api"] = []  # no sidebar on the API page
